//
//  EdenCGL.c
//
//	Implements OpenGL context setup and management via the Mac OS X CGL graphics API.
//	THIS FILE CONTAINS MAC OS X SPECIFIC CODE.
//
//  Copyright (c) 2001-2008 Philip Lamb (PRL) phil@eden.net.nz. All rights reserved.
//
//	Rev		Date		Who		Changes
//	1.0.0	2001-12-07	PRL		Initial version modified from Apple sample code for The SRMS simulator.
//	1.1.0	2004-07-11	PRL		Added ability to specify syncToVBL during setup. Added headerDoc.
//	1.2.0	2005-04-05	PRL		Changes to go with EdenEventCarbon model.
//

// @@BEGIN_EDEN_LICENSE_HEADER@@
//
//	This file is part of The Eden Library.
//
//	The Eden Library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	The Eden Library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
//
// @@END_EDEN_LICENSE_HEADER@@

// ============================================================================
//	Private includes
// ============================================================================

#include <Eden/EdenCGL.h>

#ifdef EDEN_HAVE_MACOSX_CGL

#include <OpenGL/OpenGL.h>								/* For CGL API */
//#include <OpenGL/gl.h>
//#include <OpenGL/glu.h>

// ============================================================================
//  Private defines and types
// ============================================================================

#define EDEN_CGL_DISPLAY_COUNT_MAX		8		// Who has more than this many???

#define EDEN_CGL_DEBUG			// Uncomment to build version that outputs debug info to stderr.

// ============================================================================
//	Global variables
// ============================================================================

EDEN_BOOL gEdenCGLActive = FALSE;
static CGDisplayCount gDisplayCount = 0;							// Number of displays on the system.
static CGDirectDisplayID gDisplayIDs[EDEN_CGL_DISPLAY_COUNT_MAX]; 	// CG display IDs for each display.
static CGLContextObj gDisplayCGLContexts[EDEN_CGL_DISPLAY_COUNT_MAX];	// Place to store CGLContexts if obtained.

// ============================================================================
//	Private functions
// ============================================================================

/* Report CG errors, and return to caller */
static void handleCGError(CGError err)
{
	fprintf(stderr, "CoreGraphics error: %s (%i)\n", CGLErrorString(err), (int)err);	// write err code and number to stderr
}

// 	Find out how many displays are attached to the machine and fill in 
//	gDisplayCount, and get the IDs from CoreGraphics into gDisplayIDs[].
static int EdenCGLDisplaysEnumerate(void)
{
	CGError err;
	
	// Find out how many displays we have.
	err = CGGetActiveDisplayList(0, NULL, &gDisplayCount); // see CGDirectDisplay.h
	if (err != kCGErrorSuccess) handleCGError(err);
	#ifdef EDEN_CGL_DEBUG
		fprintf(stderr, "EdenCGLDisplaysEnumerate(): Found %i displays.\n", (int)gDisplayCount);
	#endif
	if (gDisplayCount == 0) return (0);
	
	// Get the ID for each display.
	err = CGGetActiveDisplayList(EDEN_CGL_DISPLAY_COUNT_MAX, (CGDirectDisplayID *)gDisplayIDs, &gDisplayCount); // see CGDirectDisplay.h
	if (err != kCGErrorSuccess) handleCGError(err);
	
	return (gDisplayCount);
}

//  Take over a display in the system for our use.
//  Specify -1 for 'which' to capture all.
static EDEN_BOOL EdenCGLDisplayCapture(int which)
{
	CGDisplayErr err;

	if (which == -1) {
		if (CGDisplayIsCaptured(gDisplayIDs[0])) {	// Check if the main screen is captured.
			return (TRUE);
		}
		err = CGCaptureAllDisplays(); // see CGDirectDisplay.h		
	} else {
		if (CGDisplayIsCaptured(gDisplayIDs[which])) {	// Check if the screen is captured.
			return (TRUE);
		}
		err = CGDisplayCapture(gDisplayIDs[which]); // see CGDirectDisplay.h		
	}
	if (err != CGDisplayNoErr) {
		handleCGError(err);
		return (FALSE);
	}
	return (TRUE);
}

//  Give a display back to the system.
//  Specify -1 for 'which' to release all.
static EDEN_BOOL EdenCGLDisplayRelease(int which)
{
	CGDisplayErr err;
	
	if (which == -1) {
		err = CGReleaseAllDisplays();
	} else {
		err = CGDisplayRelease(gDisplayIDs[which]);
	}
	if (err != CGDisplayNoErr) {	// see CGDirectDisplay.h
		handleCGError(err);
		return (FALSE); 
	}
	return (TRUE);	
}	

// 	Utility routine: look for the specified numeric key in the Core Foundation
//	Dictionary referenced by dict.
//	Returns the value of the key as a long int, or -1 if not found.
static long GetLongFromCFDictionary(CFDictionaryRef dict, CFStringRef key)
{
	CFNumberRef numberRef;
	long value;
	
	numberRef = (CFNumberRef)CFDictionaryGetValue(dict, key);
	if (!numberRef) return (-1);
	if (!CFNumberGetValue(numberRef, kCFNumberLongType, &value)) return (-1);
	return (value);
}

//	Dissect the details of the specified display mode to stderr.
//	Display modes are reported in Core Foundation Dictionaries.
static void EdenCGLDebugReportDisplayMode(CFDictionaryRef mode)
{
	int modenum, width, height, depth, refresh;
	
	modenum = (int)GetLongFromCFDictionary(mode, kCGDisplayMode);
	width = (int)GetLongFromCFDictionary(mode, kCGDisplayWidth);
	height = (int)GetLongFromCFDictionary(mode, kCGDisplayHeight);
	depth = (int)GetLongFromCFDictionary(mode, kCGDisplayBitsPerPixel);
	refresh = (int)GetLongFromCFDictionary(mode, kCGDisplayRefreshRate);
	fprintf(stderr, "EdenCGLDebugReportDisplayMode(): Mode %i, %ix%i:%i@%i.\n",modenum,width,height,depth,refresh);
}

// ============================================================================
//	Public functions
// ============================================================================


EDEN_BOOL EdenCGLSetup(const int displayCountDesired, const EDEN_CGL_DISPLAY_MODE *displayModesDesired, const EDEN_BOOL captureAll)
{
	int i, index;
	CGDisplayErr err;
	CFDictionaryRef mode;
	CGOpenGLDisplayMask mask;
	CGLPixelFormatAttribute pixelFormatAttributes[32];	// See <OpenGL/CGLTypes.h> for values.
	CGLPixelFormatObj pixelFormat;
	long pixelFormatCount;

	if (!gDisplayCount) {
		if (EdenCGLDisplaysEnumerate() < 1) { // Count our displays, bail out if < 1.
			fprintf(stderr,"EdenCGLSetup(): No displays attached to system!\n");
			return (FALSE);
		}
	}
	
	if (displayCountDesired > gDisplayCount) {
		fprintf(stderr,"EdenCGLSetup(): %d displays requested, but only %d available.\n", displayCountDesired, gDisplayCount);
		return (FALSE);
	}

	if (captureAll) {
		if (!EdenCGLDisplayCapture(-1)) return (FALSE);				// Capture displays.
	}
	
	// Loop through requested display modes, setting up display for each one and creating a GL context on each.
	for (i = 0; i < displayCountDesired; i++) {
		
		if (displayModesDesired[i].display >= gDisplayCount) {
			fprintf(stderr,"EdenCGLSetup(): Invalid display index (%d) requested.\n", displayModesDesired[i].display);
			return (FALSE);
		}
		
		if (!captureAll) {
			if (!EdenCGLDisplayCapture(displayModesDesired[i].display)) return (FALSE);
		}
		
		// Get a suitable display mode. (The last param. is a pointer to an int in which to record
		// if we got an exact match. Not returned if NULL.) Report mode, and set it.
		if (displayModesDesired[i].refresh) {
			mode = CGDisplayBestModeForParametersAndRefreshRate(gDisplayIDs[displayModesDesired[i].display], displayModesDesired[i].depth, displayModesDesired[i].width, displayModesDesired[i].height, displayModesDesired[i].refresh, NULL);
		} else {
			mode = CGDisplayBestModeForParameters(gDisplayIDs[displayModesDesired[i].display], displayModesDesired[i].depth, displayModesDesired[i].width, displayModesDesired[i].height, NULL);
		}
#ifdef EDEN_CGL_DEBUG
		EdenCGLDebugReportDisplayMode(mode);
#endif
		err = CGDisplaySwitchToMode(gDisplayIDs[displayModesDesired[i].display], mode);
		if (err != CGDisplayNoErr) {
			fprintf(stderr, "EdenCGLSetup(): Unable to switch mode for display index %i.\n", displayModesDesired[i].display);
			handleCGError(err); // see CGDirectDisplay.h
			return (FALSE);
		}
		
		// Create an OpenGL context on the display.
		mask = CGDisplayIDToOpenGLDisplayMask(gDisplayIDs[displayModesDesired[i].display]);
		if (!mask) {
			fprintf(stderr, "EdenCGLSetup(): Unable to convert display ID to OpenGL display mask for display index %d.\n", displayModesDesired[i].display);
			return (FALSE);
		}
		index = 0;
		pixelFormatAttributes[index++] = kCGLPFADoubleBuffer;
		pixelFormatAttributes[index++] = kCGLPFAFullScreen;
		pixelFormatAttributes[index++] = kCGLPFANoRecovery;
		pixelFormatAttributes[index++] = kCGLPFAAccelerated;
		pixelFormatAttributes[index++] = kCGLPFADisplayMask;
		pixelFormatAttributes[index++] = (CGLPixelFormatAttribute)mask;
		pixelFormatAttributes[index++] = kCGLPFAColorSize;
		pixelFormatAttributes[index++] = (CGLPixelFormatAttribute)displayModesDesired[i].depth;
		pixelFormatAttributes[index++] = kCGLPFADepthSize;
		pixelFormatAttributes[index++] = (CGLPixelFormatAttribute)16;		// 16 or more bits in z-buffer.
		if (displayModesDesired[i].stereo) pixelFormatAttributes[index++] = kCGLPFAStereo;
		pixelFormatAttributes[index++] = (CGLPixelFormatAttribute)0;		// Null terminator.
		err = CGLChoosePixelFormat(pixelFormatAttributes, &pixelFormat, &pixelFormatCount);
		if (err != CGDisplayNoErr || !pixelFormat) {
			fprintf(stderr, "EdenCGLSetup(): Unable to choose pixel format for display index %i.\n", displayModesDesired[i].display);
			handleCGError(err); // see CGDirectDisplay.h
			return (FALSE);
		}
		err = CGLCreateContext(pixelFormat, NULL, &(gDisplayCGLContexts[displayModesDesired[i].display]));
		if (err != CGDisplayNoErr) {
			fprintf(stderr, "EdenCGLSetup(): Unable to create context for display index %i.\n", displayModesDesired[i].display);
			handleCGError(err); // see CGDirectDisplay.h
			return (FALSE);
		}
		CGLDestroyPixelFormat(pixelFormat);
		pixelFormat = NULL;
		
		// Set up the context, full screen, no vbl sync.
		err = CGLSetCurrentContext(gDisplayCGLContexts[displayModesDesired[i].display]);
		if (err != CGDisplayNoErr) {
			fprintf(stderr, "EdenCGLSetup(): Unable to switch to context on display index %i.\n", displayModesDesired[i].display);
			handleCGError(err); // see CGDirectDisplay.h
			return (FALSE);
		}
		err = CGLSetFullScreen(gDisplayCGLContexts[displayModesDesired[i].display]);
		if (err != CGDisplayNoErr) {
			fprintf(stderr, "EdenCGLSetup(): Unable to go to fullscreen for display index %i.\n", displayModesDesired[i].display);
			handleCGError(err); // see CGDirectDisplay.h
			return (FALSE);
		}
		if (!EdenCGLContextSetVBLSync(displayModesDesired[i].display, displayModesDesired[i].syncToVBL)) return (FALSE);
	}

	gEdenCGLActive = TRUE;
	return (TRUE);
}

EDEN_BOOL EdenCGLContextSetVBLSync(const int index, const EDEN_BOOL sync)
{
	CGError err;
	long params[1];
	
	params[0] = ( sync ? 1 : 0 );
	err = CGLSetParameter(gDisplayCGLContexts[index], kCGLCPSwapInterval, params);
	if (err != CGDisplayNoErr) {
		fprintf(stderr, "EdenCGLContextSetVBLSync(): Unable to set VBL sync to %i.\n", sync);
		handleCGError(err); // see CGDirectDisplay.h
		return (FALSE);
	}
	return (TRUE);
}


EDEN_BOOL EdenCGLContextSelect(const int index)
{
	CGError err;

	if (!gDisplayCGLContexts[index]) {
		fprintf(stderr, "EdenCGLContextSelect(): No context active on display %i.\n",index);
		return (FALSE);		
	}
	err = CGLSetCurrentContext(gDisplayCGLContexts[index]);
	if (err != CGDisplayNoErr) {
		fprintf(stderr, "EdenCGLContextSelect(): Unable to switch to context on display %i.\n",index);
		handleCGError(err); // see CGDirectDisplay.h
		return (FALSE);
	}
	return (TRUE);
}

void EdenCGLContextFlush(const int index)
{
	if (!gDisplayCGLContexts[index]) {
		fprintf(stderr,"EdenCGLContextFlush(): CGL Context %i is not active.\n",index);
		return;			// Do nothing if context not active.
	}
	CGLFlushDrawable(gDisplayCGLContexts[index]);	// Update the window.
}

EDEN_BOOL EdenCGLCleanup(void)
{
	int i;
	
	for (i = 0; i < gDisplayCount; i++)	{ // Wind up the OpenGL contexts.
		if (gDisplayCGLContexts[i]) {
			CGLSetCurrentContext(NULL);
			CGLClearDrawable(gDisplayCGLContexts[i]);
			CGLDestroyContext(gDisplayCGLContexts[i]);
			gDisplayCGLContexts[i] = NULL;
		}
	}
	if (!EdenCGLDisplayRelease(-1)) return (FALSE);		// Return the screens to the system.
	
	gEdenCGLActive = FALSE;
	return (TRUE);								// Successful cleanup.
}

#endif 	// EDEN_HAVE_MACOSX_CGL
